/*=========================================================================

  Program:   Visualization Toolkit
  Module:    vtkLinearCellExtrusionFilter.cxx

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "vtkLinearCellExtrusionFilter.h"

#include "vtkCellArray.h"
#include "vtkCellArrayIterator.h"
#include "vtkCellData.h"
#include "vtkInformation.h"
#include "vtkMergePoints.h"
#include "vtkObjectFactory.h"
#include "vtkPointData.h"
#include "vtkPolygon.h"
#include "vtkUnsignedCharArray.h"
#include "vtkUnstructuredGrid.h"

#include <array>
#include <vector>

vtkStandardNewMacro(vtkLinearCellExtrusionFilter);

//------------------------------------------------------------------------------
vtkLinearCellExtrusionFilter::vtkLinearCellExtrusionFilter()
{
  // by default process active cell scalars
  this->SetInputArrayToProcess(
    0, 0, 0, vtkDataObject::FIELD_ASSOCIATION_CELLS, vtkDataSetAttributes::SCALARS);
}

//------------------------------------------------------------------------------
int vtkLinearCellExtrusionFilter::RequestData(vtkInformation* vtkNotUsed(request),
  vtkInformationVector** inputVector, vtkInformationVector* outputVector)
{
  vtkPolyData* input = vtkPolyData::GetData(inputVector[0]);
  vtkUnstructuredGrid* output = vtkUnstructuredGrid::GetData(outputVector);

  vtkDataArray* array = nullptr;
  if (this->ScaleMode == VTK_SCALE_BY_SCALAR)
  {
    array = this->GetInputArrayToProcess(0, inputVector);
  }

  vtkCellArray* polys = input->GetPolys();
  vtkNew<vtkPoints> outputPoints;
  outputPoints->DeepCopy(input->GetPoints());

  output->SetPoints(outputPoints);

  output->GetCellData()->CopyNormalsOff();
  output->GetCellData()->ShallowCopy(input->GetCellData());

  output->GetPointData()->CopyNormalsOff();
  output->GetPointData()->CopyAllocate(input->GetPointData());

  if (this->MergeDuplicatePoints)
  {
    this->CreateDefaultLocator();
    this->Locator->SetDataSet(output);
    this->Locator->InitPointInsertion(outputPoints, outputPoints->GetBounds());

    for (vtkIdType i = 0; i < outputPoints->GetNumberOfPoints(); i++)
    {
      vtkIdType newPtId;
      this->Locator->InsertUniquePoint(outputPoints->GetPoint(i), newPtId);
      output->GetPointData()->CopyData(input->GetPointData(), i, newPtId);
    }
  }
  else
  {
    for (vtkIdType i = 0; i < outputPoints->GetNumberOfPoints(); i++)
    {
      output->GetPointData()->CopyData(input->GetPointData(), i, i);
    }
  }

  vtkDataArray* inputCellNormals = input->GetCellData()->GetNormals();
  vtkDataArray* inputPointNormals = input->GetPointData()->GetNormals();

  output->Allocate(polys->GetSize() * 2); // estimation

  std::vector<std::array<double, 3>> topPoints;
  std::vector<std::array<double, 3>> normals;
  std::vector<vtkIdType> polyhedronIds; // used for polyhedrons

  vtkIdType cellId = 0;
  auto iter = vtk::TakeSmartPointer(polys->NewIterator());
  for (iter->GoToFirstCell(); !iter->IsDoneWithTraversal(); iter->GoToNextCell(), cellId++)
  {
    vtkIdType cellSize;
    const vtkIdType* cellPoints;
    iter->GetCurrentCell(cellSize, cellPoints);

    topPoints.resize(cellSize);
    normals.resize(cellSize);
    for (vtkIdType i = 0; i < cellSize; i++)
    {
      outputPoints->GetPoint(cellPoints[i], topPoints[i].data());
    }

    for (vtkIdType i = 0; i < cellSize; i++)
    {
      if (this->UseUserVector)
      {
        normals[i][0] = this->UserVector[0];
        normals[i][1] = this->UserVector[1];
        normals[i][2] = this->UserVector[2];
      }
      else
      {
        if (this->NormalsType == VTK_CELL_NORMALS)
        {
          if (inputCellNormals)
          {
            inputCellNormals->GetTuple(cellId, normals[i].data());
          }
          else
          {
            vtkPolygon::ComputeNormal(cellSize, topPoints[0].data(), normals[i].data());
          }
        }
        else if (this->NormalsType == VTK_POINT_NORMALS)
        {
          if (inputPointNormals)
          {
            inputPointNormals->GetTuple(cellPoints[i], normals[i].data());
          }
          else
          {
            vtkErrorMacro(<< "Array of point normals does not exist!");
            return 1;
          }
        }
      }
    }

    // extrude
    double currentValue = 1.0;
    if (this->NormalsType == VTK_CELL_NORMALS)
    {
      currentValue = (array ? array->GetComponent(cellId, 0) : 1.0);
    }
    for (vtkIdType i = 0; i < cellSize; i++)
    {
      if (this->NormalsType == VTK_POINT_NORMALS)
      {
        currentValue = (array ? array->GetComponent(cellPoints[i], 0) : 1.0);
      }
      double scale = currentValue * this->ScaleFactor;
      auto& p = topPoints[i];
      p[0] += scale * normals[i][0];
      p[1] += scale * normals[i][1];
      p[2] += scale * normals[i][2];
    }

    if (cellSize == 3) // triangle => wedge
    {
      vtkIdType newPts[3];

      if (this->MergeDuplicatePoints)
      {
        this->Locator->InsertUniquePoint(topPoints[0].data(), newPts[0]);
        this->Locator->InsertUniquePoint(topPoints[1].data(), newPts[1]);
        this->Locator->InsertUniquePoint(topPoints[2].data(), newPts[2]);
      }
      else
      {
        newPts[0] = outputPoints->InsertNextPoint(topPoints[0].data());
        newPts[1] = outputPoints->InsertNextPoint(topPoints[1].data());
        newPts[2] = outputPoints->InsertNextPoint(topPoints[2].data());
      }

      vtkIdType ptsId[6] = { cellPoints[2], cellPoints[1], cellPoints[0], newPts[2], newPts[1],
        newPts[0] };

      output->InsertNextCell(VTK_WEDGE, 6, ptsId);

      for (int i = 0; i < cellSize; ++i)
      {
        output->GetPointData()->CopyData(input->GetPointData(), cellPoints[i], newPts[i]);
      }
    }
    else if (cellSize == 4) // quad => hexahedron
    {
      vtkIdType newPts[4];

      if (this->MergeDuplicatePoints)
      {
        this->Locator->InsertUniquePoint(topPoints[0].data(), newPts[0]);
        this->Locator->InsertUniquePoint(topPoints[1].data(), newPts[1]);
        this->Locator->InsertUniquePoint(topPoints[2].data(), newPts[2]);
        this->Locator->InsertUniquePoint(topPoints[3].data(), newPts[3]);
      }
      else
      {
        newPts[0] = outputPoints->InsertNextPoint(topPoints[0].data());
        newPts[1] = outputPoints->InsertNextPoint(topPoints[1].data());
        newPts[2] = outputPoints->InsertNextPoint(topPoints[2].data());
        newPts[3] = outputPoints->InsertNextPoint(topPoints[3].data());
      }

      vtkIdType ptsId[8] = { cellPoints[3], cellPoints[2], cellPoints[1], cellPoints[0], newPts[3],
        newPts[2], newPts[1], newPts[0] };

      output->InsertNextCell(VTK_HEXAHEDRON, 8, ptsId);

      for (int i = 0; i < cellSize; ++i)
      {
        output->GetPointData()->CopyData(input->GetPointData(), cellPoints[i], newPts[i]);
      }
    }
    else // generic case => polyhedron
    {
      polyhedronIds.resize(2 * (cellSize + 1) + cellSize * 5);

      vtkIdType* topFace = polyhedronIds.data();
      vtkIdType* baseFace = topFace + cellSize + 1;

      topFace[0] = cellSize;
      baseFace[0] = cellSize;
      for (vtkIdType i = 0; i < cellSize; i++)
      {
        if (this->MergeDuplicatePoints)
        {
          this->Locator->InsertUniquePoint(topPoints[i].data(), topFace[i + 1]);
        }
        else
        {
          topFace[i + 1] = outputPoints->InsertNextPoint(topPoints[i].data());
        }
        baseFace[i + 1] = cellPoints[cellSize - i - 1];

        output->GetPointData()->CopyData(input->GetPointData(), cellPoints[i], topFace[i]);
      }

      for (vtkIdType i = 0; i < cellSize; i++)
      {
        vtkIdType* currentSide = polyhedronIds.data() + 2 * (cellSize + 1) + 5 * i;
        currentSide[0] = 4;
        currentSide[1] = topFace[1 + (i + 1) % cellSize];
        currentSide[2] = topFace[1 + i];
        currentSide[3] = cellPoints[i];
        currentSide[4] = cellPoints[(i + 1) % cellSize];
      }

      output->InsertNextCell(VTK_POLYHEDRON, cellSize + 2, polyhedronIds.data());
    }

    if (cellId % 1000 == 0)
    {
      this->UpdateProgress(cellId / static_cast<double>(polys->GetNumberOfCells()));
    }
  }

  output->Squeeze();

  this->UpdateProgress(1.0);

  return 1;
}

//------------------------------------------------------------------------------
void vtkLinearCellExtrusionFilter::CreateDefaultLocator()
{
  if (!this->Locator)
  {
    this->Locator = vtkSmartPointer<vtkMergePoints>::New();
  }
}

//------------------------------------------------------------------------------
void vtkLinearCellExtrusionFilter::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os, indent);

  os << indent << "Normals Type: ";
  if (this->NormalsType == VTK_CELL_NORMALS)
  {
    os << "Extrude along cell normals" << endl;
  }
  else if (this->NormalsType == VTK_POINT_NORMALS)
  {
    os << "Extrude along point normals" << endl;
  }

  os << indent << "Scale Mode: ";
  if (this->NormalsType == VTK_CELL_NORMALS)
  {
    os << "Scale by scalar" << endl;
  }
  else if (this->NormalsType == VTK_POINT_NORMALS)
  {
    os << "Data scaling is turned off" << endl;
  }

  os << indent << "ScaleFactor: " << this->ScaleFactor << endl
     << indent << "UserVector: " << this->UserVector[0] << " " << this->UserVector[1] << " "
     << this->UserVector[2] << endl
     << indent << "UseUserVector: " << (this->UseUserVector ? "ON" : "OFF") << endl
     << indent << "MergeDuplicatePoints: " << (this->MergeDuplicatePoints ? "ON" : "OFF") << endl;
}

//------------------------------------------------------------------------------
int vtkLinearCellExtrusionFilter::FillOutputPortInformation(
  int vtkNotUsed(port), vtkInformation* info)
{
  info->Set(vtkDataObject::DATA_TYPE_NAME(), "vtkUnstructuredGrid");
  return 1;
}
